/**
 * Copyright (c) 2021 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "turn.h"

pokerturn_t pokerTurnGet(poker_t *poker, uint8_t playerIndex) {
  pokerplayer_t *player;
  float t, confidence, cardWeight, handWeight;
  float betWeight, potWeight,  inWeight, roundWeight, bet;
  card_t bestCard;
  uint8_t i, bestCardNumber;
  int32_t chips;

  player = poker->players + playerIndex;

  // Can the player do anything?
  if(player->state & (POKER_PLAYER_STATE_FOLDED | POKER_PLAYER_STATE_OUT)) {
    return pokerTurnOut(poker, playerIndex);
  }

  // Now we basically need to determine the AI's action here. The AI really only
  // needs to do one thing, decide whether or not they want to Bet or Not.
  // In future I may make checking optional, but for now there's really no 
  // reason not to.

  // Get the current winning hand.
  pokerplayerwinning_t winning;
  pokerWinnerPlayerGet(&poker->dealer, player, &winning);

  // Now try and figure out how good the hand is
  handWeight = 0;
  for(i = 0; i < winning.setSize; i++) {
    bestCard = winning.set[i];
    handWeight += pokerWinnerGetCardWeight(bestCard);
  }
  handWeight = handWeight / winning.setSize;

  bestCard = pokerWinnerGetBestCard(winning.set, winning.setSize);
  bestCardNumber = cardGetNumber(bestCard);
  confidence = pokerWinnerGetTypeConfidence(winning.type);
  cardWeight = pokerWinnerGetCardWeight(bestCard);
  
  // Now figure out what the risk of betting is.
  inWeight = (float)player->currentBet / (float)player->chips;
  roundWeight = (
    ((float)poker->dealer.cardsFacing) / ((float)POKER_DEALER_HAND_SIZE)
  );
  betWeight = (float)(
    poker->bet.currentBet - player->currentBet
  ) / (float)player->chips;
  potWeight = poker->bet.pot / (float)player->chips;

  
  // We now need to act on these values.
  printf("Card Weights - %.4f, %.4f, %.4f, %.4f\n", confidence, cardWeight, handWeight);
  printf("Money Weights - %.4f, %.4f, %.4f, %.4f\n", inWeight, roundWeight, betWeight, potWeight);

  // Chip weight, 1 is all in.
  bet = 0;

  if(roundWeight == 0) { 
    // First Round

    if(betWeight < 0.1) {
      // The bet is "low risk" to me. (<10% of my chips), play a bit loose.
      // Take any card with 70% weighting and/or any hand with a pair or better.
      // if(cardWeight > 0.7 || confidence >= POKER_WINNNIG_CONFIDENCE_PAIR) {
        // bet = confidence * 0.7f;
      // } else if(cardWeight >= 0.5 && betWeight > 0) {
      // }
      bet = (1.0f - mathMin(betWeight*10, 0.5f)) * (cardWeight * confidence * 2.0f);
    } else if(betWeight < 0.4) {
      // Bet is "medium" risk to me (less than 40% of my chips)
      // Bet on any hand that has a good chance.
      if(confidence >= POKER_WINNNIG_CONFIDENCE_PAIR) {
        bet = (confidence * mathMax(cardWeight, betWeight));
      }
    } else if(betWeight < 0.75) {
      // Bet is "high" risk to me.
      if(cardWeight > 0.6 && confidence >= POKER_WINNNIG_CONFIDENCE_PAIR) {
        bet = (confidence * cardWeight);
      }
    } else {
      // Bet is likely my last hand
    }
  } else {
    // Not first round.
    if(confidence <= POKER_WINNNIG_CONFIDENCE_PAIR) {
      bet = (confidence * cardWeight) / betWeight;
    } else if(confidence <= POKER_WINNNIG_CONFIDENCE_THREE_OF_A_KIND) {
      bet = (confidence * (cardWeight / handWeight)) / betWeight;
    } else {
      bet = confidence * (cardWeight / handWeight) * 2;
    }
  }

  
  printf("Bet %.4f\n\n", bet);

  // Now determine chips based on bet.
  if(bet >= 0.75) {
    chips = (int32_t)(bet * (float)player->chips);
    chips = mathMin(chips, player->chips);
    return pokerTurnRaise(poker, playerIndex, chips);
  } else if(bet >= 0.25f) {
    return pokerTurnCall(poker, playerIndex);
  }

  
  // Well, now they have decided to "not bet", or really they think that their
  // cards are not worth the effort. But if they CAN check, they will.
  if(pokerTurnCanPlayerCheck(poker, playerIndex)) {
    return pokerTurnCheck(poker, playerIndex);
  }

  // Nothing worth doing, so let's just fold here.
  return pokerTurnFold(poker, playerIndex);
}

void pokerTurnAction(poker_t *poker, pokerplayer_t *player, pokerturn_t *turn) {
  switch(turn->type) {
    case POKER_TURN_TYPE_BET:
    case POKER_TURN_TYPE_CALL:
    case POKER_TURN_TYPE_ALL_IN:
      pokerBetPlayer(&poker->bet, player, turn->chips);
      break;

    case POKER_TURN_TYPE_CHECK:
      pokerBetPlayer(&poker->bet, player, 0);
      break;

    case POKER_TURN_TYPE_FOLD:
      player->state |= POKER_PLAYER_STATE_FOLDED;
      break;
  }
  
  player->state |= POKER_PLAYER_STATE_ROUND_MOVE;
}

pokerturn_t pokerTurnOut(poker_t *poker, uint8_t player) {
  pokerturn_t turn;
  turn.type = POKER_TURN_TYPE_OUT;
  return turn;
}

pokerturn_t pokerTurnFold(poker_t *poker, uint8_t player) {
  pokerturn_t turn;
  turn.type = POKER_TURN_TYPE_FOLD;
  return turn;
}

pokerturn_t pokerTurnCheck(poker_t *poker, uint8_t player) {
  pokerturn_t turn;
  turn.type = POKER_TURN_TYPE_CHECK;
  return turn;
}

pokerturn_t pokerTurnCall(poker_t *poker, uint8_t playerIndex) {
  pokerturn_t turn;
  pokerplayer_t *player;
  player = poker->players + playerIndex;
  turn = pokerTurnRaise(
    poker, playerIndex, poker->bet.currentBet - player->currentBet
  );
  return turn;
}

pokerturn_t pokerTurnRaise(poker_t *poker, uint8_t playerIndex, int32_t chips) {
  pokerturn_t turn;
  pokerplayer_t *player;
  player = poker->players + playerIndex;

  if(chips < poker->bet.currentBet) chips = poker->bet.currentBet;

  if(player->chips <= chips) {
    turn.chips = player->chips;
    turn.type = POKER_TURN_TYPE_ALL_IN;
  } else {
    turn.chips = chips;
    turn.type = POKER_TURN_TYPE_BET;

    if(chips == poker->bet.currentBet) {
      turn.type = POKER_TURN_TYPE_CALL;
    }

    if(chips == 0) {
      turn.type = POKER_TURN_TYPE_CALL;
    }
  }

  return turn;
}

bool pokerTurnCanPlayerCheck(poker_t *poker, uint8_t playerIndex) {
  return (poker->players + playerIndex)->currentBet >= poker->bet.currentBet;
}